name = "scanapi"

import click
import logging
import yaml

from scanapi.errors import (
    EmptyConfigFileError,
    FileFormatNotSupportedError,
    InvalidKeyError,
)
from scanapi.tree import EndpointNode
from scanapi.reporter import Reporter
from scanapi.settings import settings
from scanapi.config_loader import load_config_file


@click.command()
@click.option("-s", "--spec-path", "spec_path", type=click.Path(exists=True))
@click.option("-o", "--output-path", "output_path", type=click.Path())
@click.option("-c", "--config-path", "config_path", type=click.Path(exists=True))
@click.option(
    "-r", "--reporter", "reporter", type=click.Choice(["console", "markdown", "html"])
)
@click.option("-t", "--template", "template", type=click.Path(exists=True))
@click.option(
    "--log-level",
    "log_level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO",
)
def scan(spec_path, output_path, config_path, reporter, template, log_level):
    """Automated Testing and Documentation for your REST API."""
    logging.basicConfig(level=log_level)
    logger = logging.getLogger(__name__)

    settings.save_preferences(
        **{
            "spec_path": spec_path,
            "output_path": output_path,
            "config_path": config_path,
            "reporter": reporter,
            "template": template,
        }
    )

    spec_path = settings["spec_path"]

    try:
        api_spec = load_config_file(spec_path)
    except FileNotFoundError as e:
        error_message = f"Could not find API spec file: {spec_path}. {str(e)}"
        logger.error(error_message)
        return
    except EmptyConfigFileError as e:
        error_message = f"API spec file is empty. {str(e)}"
        logger.error(error_message)
        return
    except (yaml.YAMLError, FileFormatNotSupportedError) as e:
        logger.error(e)
        return

    try:
        root_node = EndpointNode(api_spec["api"])
        Reporter(
            settings["output_path"], settings["reporter"], settings["template"]
        ).write(root_node.run())
    except InvalidKeyError as e:
        error_message = "Error loading API spec."
        error_message = "{} {}".format(error_message, str(e))
        logger.error(error_message)
        return



from functools import singledispatch
import logging

from scanapi.evaluators.string_evaluator import StringEvaluator

logger = logging.getLogger(__name__)


class SpecEvaluator:
    def __init__(self, endpoint, vars={}):
        self.endpoint = endpoint
        self.registry = {}
        self.update(vars)

    def evaluate(self, element):
        return evaluate(element, self)

    def update(self, vars, extras=None, preevaluate=False):
        if preevaluate:
            values = {key: evaluate(value, extras) for key, value in vars.items()}
            self.registry.update(values)
        else:
            self.registry.update(vars)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __repr__(self):
        return self.registry.__repr__()

    def __getitem__(self, key):
        if key in self:
            return self.registry[key]

        if key in self.endpoint.parent.vars:
            return self.endpoint.parent.vars[key]

        raise KeyError(key)

    def __contains__(self, key):
        return key in self.registry


@singledispatch
def evaluate(expression, vars):
    return expression


@evaluate.register(str)
def _evaluate_str(element, vars):
    return StringEvaluator.evaluate(element, vars)


@evaluate.register(dict)
def _evaluate_dict(element, vars):
    return {key: evaluate(value, vars) for key, value in element.items()}


@evaluate.register(list)
@evaluate.register(tuple)
def _evaluate_collection(elements, vars):
    return [evaluate(item, vars) for item in elements]



import logging
import os
import re
import sys

from scanapi.errors import BadConfigurationError, InvalidPythonCodeError
from scanapi.evaluators.code_evaluator import CodeEvaluator

logger = logging.getLogger(__name__)


class StringEvaluator:
    variable_pattern = re.compile(
        r"(?P<something_before>\w*)(?P<start>\${)(?P<variable>\w*)(?P<end>})(?P<something_after>\w*)"
    )  # ${<variable>}

    @classmethod
    def evaluate(cls, sequence, vars):
        try:
            sequence = cls._evaluate_env_var(sequence)
        except BadConfigurationError as e:
            logger.error(e)
            sys.exit()

        sequence = cls._evaluate_custom_var(sequence, vars)

        try:
            return CodeEvaluator.evaluate(sequence, vars)
        except InvalidPythonCodeError as e:
            logger.error(e)
            sys.exit()

    @classmethod
    def _evaluate_env_var(cls, sequence):
        matches = cls.variable_pattern.finditer(sequence)

        if not matches:
            return sequence

        for match in matches:
            variable_name = match.group("variable")

            if any(letter.islower() for letter in variable_name):
                continue

            try:
                variable_value = os.environ[variable_name]
            except KeyError as e:
                raise BadConfigurationError(e)

            sequence = cls.replace_var_with_value(
                sequence, match.group(), variable_value
            )

        return sequence

    @classmethod
    def _evaluate_custom_var(cls, sequence, vars):
        matches = cls.variable_pattern.finditer(sequence)

        if not matches:
            return sequence

        for match in matches:
            variable_name = match.group("variable")

            if variable_name.isupper():
                continue

            if not vars.get(variable_name):
                continue

            variable_value = vars.get(variable_name)

            sequence = cls.replace_var_with_value(
                sequence, match.group(), variable_value
            )

        return sequence

    @classmethod
    def replace_var_with_value(cls, sequence, variable, variable_value):
        variable = re.escape(variable)
        return re.sub(variable, variable_value, sequence)



import pytest
import jinja2
from scanapi.reporter import Reporter

fake_responses = [
    {"status_code": 200, "request": {"method": "GET", "url": "http://test.com"}}
]


class TestConsoleReport:
    @pytest.fixture
    def mocked_print(self, mocker):
        return mocker.patch("builtins.print")

    def test_should_print(self, mocker, mocked_print):
        console_reporter = Reporter(None, "console")
        console_reporter.write(fake_responses)

        expected_content = "\n".join(
            (
                "ScanAPI Report: Console",
                "=======================",
                "",
                "GET http://test.com - 200",
                "",
            )
        )

        mocked_print.assert_called_once_with(f"\n{expected_content}")


class TestReporterOtherThanConsole:
    @pytest.fixture
    def mocked__render_content(self, mocker):
        return mocker.patch("scanapi.reporter.Reporter._render_content")

    @pytest.fixture
    def mocked_open(self, mocker):
        mock = mocker.mock_open()
        mocker.patch("builtins.open", mock)
        return mock

    @pytest.mark.parametrize("reporter_type", ["html", "markdown"])
    def test_should_write_to_default_output(
        self, reporter_type, mocker, mocked__render_content, mocked_open
    ):
        mocked__render_content.return_value = "ScanAPI Report"
        reporter = Reporter(None, reporter_type)
        reporter.write(fake_responses)

        file_extension = {"html": "html", "markdown": "md"}[reporter_type]
        mocked_open.assert_called_once_with(
            f"scanapi-report.{file_extension}", "w", newline="\n"
        )
        mocked_open().write.assert_called_once_with("ScanAPI Report")

    def test_should_write_to_custom_output(
        self, mocker, mocked__render_content, mocked_open
    ):
        mocked__render_content.return_value = "ScanAPI Report"
        reporter = Reporter("./custom/report-output.html", "html")
        reporter.write(fake_responses)

        mocked_open.assert_called_once_with(
            "./custom/report-output.html", "w", newline="\n"
        )
        mocked_open().write.assert_called_once_with("ScanAPI Report")

    def test_should_handle_custom_templates(
        self, mocker, mocked__render_content, mocked_open
    ):
        mocked__render_content.return_value = "ScanAPI Report"
        reporter = Reporter(None, None, "my-template.html")
        reporter.write(fake_responses)

        mocked_open.assert_called_once_with("scanapi-report", "w", newline="\n")
        mocked_open().write.assert_called_once_with("ScanAPI Report")



import os
import pytest
import requests

from scanapi.errors import BadConfigurationError, InvalidPythonCodeError
from scanapi.evaluators import CodeEvaluator, SpecEvaluator, StringEvaluator


class TestCodeEvaluator:
    class TestEvaluate:
        class TestWhenDoesNotMatchThePattern:
            test_data = ["no code", "${CODE}", "${code}", "{{code}}"]

            @pytest.mark.parametrize("sequence", test_data)
            def test_should_return_sequence(self, sequence):
                assert CodeEvaluator.evaluate(sequence, {}) == sequence

        class TestWhenMatchesThePattern:
            test_data = [("${{1 + 1}}", "2"), ("${{'hi'*4}}", "hihihihi")]

            @pytest.mark.parametrize("sequence, expected", test_data)
            def test_should_return_evaluated_code(self, sequence, expected):
                assert CodeEvaluator.evaluate(sequence, {}) == expected

            class TestWhenCodeContainsPreSavedResponse:
                @pytest.fixture
                def response(self, requests_mock):
                    requests_mock.get("http://test.com", text="abcde")
                    return requests.get("http://test.com")

                test_data = [
                    ("${{response.text}}", "abcde"),
                    ("${{response.status_code}}", "200"),
                    ("${{response.text + 'xpto'}}", "abcdexpto"),
                    ("${{'xpto' + response.text}}", "xptoabcde"),
                    ("${{1+1}}", "2"),
                ]

                @pytest.mark.parametrize("sequence, expected", test_data)
                def test_should_return_evaluated_code(
                    self, sequence, expected, response
                ):
                    assert (
                        CodeEvaluator.evaluate(sequence, {"response": response})
                        == expected
                    )

            class TestWhenCodeBreaks:
                def test_should_raises_invalid_python_code_error(self):
                    with pytest.raises(InvalidPythonCodeError) as excinfo:
                        CodeEvaluator.evaluate("${{1/0}}", {})

                    assert (
                        str(excinfo.value)
                        == "Invalid Python code defined in the API spec: division by zero"
                    )



import pytest
import os

from scanapi.settings import settings, DEFAULT_CONFIG_PATH


class TestSettings:
    class TestInit:
        def test_should_init_with_default_values(self):
            assert settings["spec_path"] == "api.yaml"
            assert settings["reporter"] == "markdown"
            assert settings["output_path"] is None
            assert settings["template"] is None

    class TestSavePreferences:
        @pytest.fixture
        def mock_save_click_preferences(self, mocker):
            return mocker.patch("scanapi.settings.save_click_preferences")

        @pytest.fixture
        def mock_save_config_file_preferences(self, mocker):
            return mocker.patch("scanapi.settings.save_config_file_preferences")

        class TestWhenConfigPathIsInClickPreferences:
            def test_should_pass_config_path(
                self, mock_save_click_preferences, mock_save_config_file_preferences
            ):
                preferences = {"config_path": "foo.yaml"}
                settings.save_preferences(**preferences)
                mock_save_config_file_preferences.assert_called_with("foo.yaml")
                mock_save_click_preferences.assert_called_with(**preferences)

        class TestWhenConfigPathIsNotInClickPreferences:
            def test_should_pass_config_path_as_none(
                self, mock_save_click_preferences, mock_save_config_file_preferences
            ):
                preferences = {"foo": "bar"}
                settings.save_preferences(**preferences)
                mock_save_config_file_preferences.assert_called_with(None)
                mock_save_click_preferences.assert_called_with(**preferences)

    class TestSaveClickPreferences:
        def test_should_clean_and_save_preferences(self):
            settings.save_click_preferences(
                **{
                    "spec_path": "path/spec-path",
                    "reporter": None,
                    "output_path": "path/output-path",
                    "template": None,
                    "config_path": "path/config-path",
                }
            )

            assert settings == {
                "spec_path": "path/spec-path",
                "reporter": "markdown",
                "output_path": "path/output-path",
                "template": None,
                "config_path": "path/config-path",
            }

    class TestSaveConfigFilePreferences:
        @pytest.fixture(autouse=True)
        def define_settings(self):
            settings["spec_path"] = "path/spec-path"
            settings["reporter"] = "markdown"
            settings["output_path"] = "path/output-path"
            settings["template"] = None
            settings["config_path"] = "path/config-path"

        class TestWithoutCustomConfigPath:
            class TestWithDefaultConfigFile:
                def test_should_save_preferences(self):
                    with open(DEFAULT_CONFIG_PATH, "w") as out_file:
                        out_file.write("reporter: html")

                    settings.save_config_file_preferences()
                    assert settings["reporter"] == "html"

                    os.remove(DEFAULT_CONFIG_PATH)

            class TestWithoutDefaultConfigFile:
                def test_should_not_change_preferences(self):
                    settings.save_config_file_preferences()
                    assert settings["reporter"] == "markdown"

        class TestWithCustomConfigPath:
            class TestWithConfigFile:
                def test_should_save_preferences(self):
                    config_path = "my_config_file.yaml"
                    with open(config_path, "w") as out_file:
                        out_file.write("reporter: html")

                    settings.save_config_file_preferences(config_path)
                    assert settings["reporter"] == "html"

                    os.remove(config_path)

            class TestWithoutConfigFile:
                def test_should_raise_exception(self):
                    with pytest.raises(FileNotFoundError) as excinfo:
                        config_path = "invalid/my_config_file.yaml"
                        settings.save_config_file_preferences(config_path)

                    assert (
                        str(excinfo.value)
                        == "[Errno 2] No such file or directory: 'invalid/my_config_file.yaml'"
                    )



class MalformedSpecError(Exception):
    pass


class HTTPMethodNotAllowedError(MalformedSpecError):
    """Raised when the HTTP method in the API spec is invalid"""

    def __init__(self, method, allowed_methos, *args):
        message = (
            f"HTTP method not supported: {method}. Supported methods: {allowed_methos}."
        )
        super(HTTPMethodNotAllowedError, self).__init__(message, *args)


class APIKeyMissingError(MalformedSpecError):
    """Raised when 'api' key is not specified at root scope in the API spec"""

    def __init__(self, *args):
        message = "Missing api 'key' at root scope in the API spec"
        super(APIKeyMissingError, self).__init__(message, *args)


class InvalidKeyError(MalformedSpecError):
    """Raised when an invalid key is specified in the API spec"""

    def __init__(self, key, scope, available_keys, *args):
        message = f"Invalid key '{key}' at '{scope}' scope. Available keys are: {available_keys}"
        super(InvalidKeyError, self).__init__(message, *args)


class InvalidPythonCodeError(MalformedSpecError):
    """Raised when python code defined in the API spec raises an error"""

    def __init__(self, error_message, *args):
        error_message = f"Invalid Python code defined in the API spec: {error_message}"
        super(InvalidPythonCodeError, self).__init__(error_message, *args)


class BadConfigurationError(Exception):
    """Raised when an environment variable was not set or was badly configured"""

    def __init__(self, env_var, *args):
        super(BadConfigurationError, self).__init__(
            f"{env_var} environment variable not set or badly configured", *args
        )


class EmptyConfigFileError(Exception):
    """Raised when the Config File loaded is empty"""

    def __init__(self, file_path, *args):
        message = f"File '{file_path}' is empty."
        super(EmptyConfigFileError, self).__init__(message, *args)


class FileFormatNotSupportedError(Exception):
    """Raised when the Config File format is not supported"""

    def __init__(self, file_format, file_path, *args):
        message = f"The format {file_format} is not supported. Supported formats: '.yaml', '.yml', '.json'. File path: '{file_path}'."
        super(FileFormatNotSupportedError, self).__init__(message, *args)



import pytest
import requests

from scanapi.errors import InvalidKeyError
from scanapi.utils import (
    _hide,
    _override_info,
    hide_sensitive_info,
    join_urls,
    validate_keys,
)


@pytest.fixture
def response(requests_mock):
    requests_mock.get("http://test.com", text="data")
    return requests.get("http://test.com")


class TestJoinUrls:
    test_data = [
        (
            "http://demo.scanapi.dev/api/",
            "health",
            "http://demo.scanapi.dev/api/health",
        ),
        (
            "http://demo.scanapi.dev/api/",
            "/health",
            "http://demo.scanapi.dev/api/health",
        ),
        (
            "http://demo.scanapi.dev/api/",
            "/health/",
            "http://demo.scanapi.dev/api/health/",
        ),
        ("http://demo.scanapi.dev/api", "health", "http://demo.scanapi.dev/api/health"),
        (
            "http://demo.scanapi.dev/api",
            "/health",
            "http://demo.scanapi.dev/api/health",
        ),
        (
            "http://demo.scanapi.dev/api",
            "/health/",
            "http://demo.scanapi.dev/api/health/",
        ),
        (
            "",
            "http://demo.scanapi.dev/api/health/",
            "http://demo.scanapi.dev/api/health/",
        ),
        (
            "http://demo.scanapi.dev/api/health/",
            "",
            "http://demo.scanapi.dev/api/health/",
        ),
        ("", "", ""),
    ]

    @pytest.mark.parametrize("url_1, url_2, expected", test_data)
    def test_build_url_properly(self, url_1, url_2, expected):
        assert join_urls(url_1, url_2) == expected


class TestValidateKeys:
    class TestThereIsAnInvalidKey:
        def test_should_raise_an_exception(self):
            keys = ["key1", "key2"]
            available_keys = ("key1", "key3")
            scope = "endpoint"

            with pytest.raises(InvalidKeyError) as excinfo:
                validate_keys(keys, available_keys, scope)

            assert (
                str(excinfo.value)
                == "Invalid key 'key2' at 'endpoint' scope. Available keys are: ('key1', 'key3')"
            )

    class TestThereIsNotAnInvalidKeys:
        def test_should_not_raise_an_exception(self):
            keys = ["key1"]
            available_keys = ("key1", "key3")
            scope = "endpoint"

            validate_keys(keys, available_keys, scope)


class TestHideSensitiveInfo:
    @pytest.fixture
    def mock__hide(self, mocker):
        return mocker.patch("scanapi.utils._hide")

    test_data = [
        ({}, {}, {}),
        ({"report": {"abc": "def"}}, {}, {}),
        ({"report": {"hide-request": {"headers": ["abc"]}}}, {"headers": ["abc"]}, {}),
        ({"report": {"hide-response": {"headers": ["abc"]}}}, {}, {"headers": ["abc"]}),
    ]

    @pytest.mark.parametrize("settings, request_settings, response_settings", test_data)
    def test_calls__hide(
        self,
        settings,
        request_settings,
        response_settings,
        mocker,
        response,
        mock__hide,
    ):
        mocker.patch("scanapi.utils.settings", settings)
        hide_sensitive_info(response)

        calls = [
            mocker.call(response.request, request_settings),
            mocker.call(response, response_settings),
        ]

        mock__hide.assert_has_calls(calls)


class TestHide:
    @pytest.fixture
    def mock__override_info(self, mocker):
        return mocker.patch("scanapi.utils._override_info")

    test_data = [
        ({}, []),
        ({"headers": ["abc", "def"]}, [("headers", "abc"), ("headers", "def")]),
        ({"headers": ["abc"]}, [("headers", "abc")]),
    ]

    @pytest.mark.parametrize("settings, calls", test_data)
    def test_calls__override_info(
        self, settings, calls, mocker, response, mock__override_info
    ):
        _hide(response, settings)
        calls = [mocker.call(response, call[0], call[1]) for call in calls]

        mock__override_info.assert_has_calls(calls)


class TestOverrideInfo:
    def test_overrides(self, response):
        response.headers = {"abc": "123"}
        http_attr = "headers"
        secret_field = "abc"

        _override_info(response, http_attr, secret_field)

        assert response.headers["abc"] == "<sensitive_information>"

    def test_when_http_attr_is_not_allowed(self, response, mocker):
        mocker.patch("scanapi.utils.ALLOWED_ATTRS_TO_HIDE", ["body"])
        response.headers = {"abc": "123"}
        http_attr = "headers"
        secret_field = "abc"

        _override_info(response, http_attr, secret_field)

        assert response.headers["abc"] == "123"

    def test_when_http_attr_does_not_have_the_field(self, response, mocker):
        mocker.patch("scanapi.utils.ALLOWED_ATTRS_TO_HIDE", ["body"])
        response.headers = {"abc": "123"}
        http_attr = "headers"
        secret_field = "def"

        _override_info(response, http_attr, secret_field)

        assert response.headers == {"abc": "123"}



#!/usr/bin/env python3
import logging
import curlify

from jinja2 import Environment, FileSystemLoader, PackageLoader, select_autoescape

from scanapi.settings import settings

logger = logging.getLogger(__name__)


class Reporter:
    def __init__(self, output_path, reporter, template=None):
        self.output_path = output_path
        self.reporter = reporter
        self.template = template

    def write(self, responses):
        logger.info("Writing documentation")

        if self.template:
            loader = FileSystemLoader(searchpath="./")
            template_path = self.template
        else:
            loader = PackageLoader("scanapi", "templates")
            template_path = f"{self.reporter}.jinja"

        content = self._render_content(loader, template_path, responses)

        if self.reporter == "console":
            print(f"\n{content}")
            return

        if self.output_path is None:
            outputs = {"html": "scanapi-report.html", "markdown": "scanapi-report.md"}
            self.output_path = outputs.get(self.reporter) or "scanapi-report"

        with open(self.output_path, "w", newline="\n") as doc:
            doc.write(content)

        logger.info("The documentation was generated successfully.")
        logger.info(f"It is available at {self.output_path}")

    def _render_content(self, loader, template_path, responses):
        env = Environment(loader=loader)
        env.filters["curlify"] = curlify.to_curl
        chosen_template = env.get_template(template_path)
        return chosen_template.render(responses=responses)



import pytest
import yaml

from scanapi.errors import EmptyConfigFileError, FileFormatNotSupportedError
from scanapi.config_loader import load_config_file


class TestLoadConfigFile:
    def test_should_load(self):
        data = load_config_file("tests/data/api.yaml")
        assert data == {
            "api": {
                "endpoints": [
                    {
                        "name": "scanapi-demo",
                        "path": "${BASE_URL}",
                        "requests": [{"name": "health", "path": "/health/"}],
                    }
                ]
            }
        }

    class TestLoadJson:
        def test_should_load(self):
            data = load_config_file("tests/data/jsonfile.json")
            assert data == {
                "api": {
                    "endpoints": [
                        {
                            "name": "scanapi-demo",
                            "path": "${BASE_URL}",
                            "requests": [{"name": "health", "path": "/health/"}],
                        }
                    ]
                }
            }

    class TestWhenFileDoesNotExist:
        def test_should_raise_exception(self):

            with pytest.raises(FileNotFoundError) as excinfo:
                load_config_file("invalid/path.yaml")

            assert (
                str(excinfo.value)
                == "[Errno 2] No such file or directory: 'invalid/path.yaml'"
            )

    class TestWhenFileIsEmpty:
        def test_should_raise_exception(self):
            with pytest.raises(EmptyConfigFileError) as excinfo:
                load_config_file("tests/data/empty.yaml")

            assert str(excinfo.value) == "File 'tests/data/empty.yaml' is empty."

    class TestWhenFileFormatIsNotSupported:
        def test_should_raise_exception(self):
            with pytest.raises(FileFormatNotSupportedError) as excinfo:
                load_config_file("tests/data/not_supported_format.txt")

            assert (
                str(excinfo.value)
                == "The format .txt is not supported. Supported formats: '.yaml', '.yml', '.json'. "
                "File path: 'tests/data/not_supported_format.txt'."
            )

    class TestWhenIncludeFileFormatIsNotSupported:
        def test_should_raise_exception(self):
            with pytest.raises(FileFormatNotSupportedError) as excinfo:
                load_config_file("tests/data/api_wrong_format_include.yaml")

            assert (
                str(excinfo.value)
                == "The format .txt is not supported. Supported formats: '.yaml', '.yml', '.json'. "
                "File path: 'tests/data/include.txt'."
            )

    class TestWhenIncludeFileDoesNotExist:
        def test_should_raise_exception(self):

            with pytest.raises(FileNotFoundError) as excinfo:
                load_config_file("tests/data/api_invalid_path_include.yaml")

            assert "[Errno 2] No such file or directory: " in str(excinfo.value)
            assert "tests/data/invalid_path/include.yaml'" in str(excinfo.value)



import pytest

from scanapi.errors import HTTPMethodNotAllowedError
from scanapi.tree import EndpointNode, RequestNode


class TestRequestNode:
    @pytest.fixture
    def mock_evaluate(self, mocker):
        mock_func = mocker.patch("scanapi.tree.request_node.SpecEvaluator.evaluate")
        mock_func.return_value = ""

        return mock_func

    class TestHTTPMethod:
        def test_when_request_has_method(self):
            request = RequestNode({"method": "put"}, endpoint=EndpointNode({}))
            assert request.http_method == "PUT"

        def test_when_request_has_no_method(self):
            request = RequestNode({}, endpoint=EndpointNode({}))
            assert request.http_method == "GET"

        def test_when_method_is_invalid(self):
            request = RequestNode({"method": "xxx"}, endpoint=EndpointNode({}))
            with pytest.raises(HTTPMethodNotAllowedError) as excinfo:
                request.http_method

            assert (
                str(excinfo.value) == "HTTP method not supported: XXX. "
                "Supported methods: ('GET', 'POST', 'PUT', 'PATCH', 'DELETE')."
            )

    class TestName:
        def test_when_request_has_name(self):
            request = RequestNode({"name": "list-users"}, endpoint=EndpointNode({}))
            assert request.name == "list-users"

        @pytest.mark.skip("it should validate mandatory `name` key before")
        def test_when_request_has_no_name(self):
            request = RequestNode({}, endpoint=EndpointNode({}))

    class TestFullPathUrl:
        def test_request_with_no_path(self):
            base_path = "http://foo.com/"
            request = RequestNode({}, endpoint=EndpointNode({"path": base_path}))
            assert request.full_url_path == base_path

        def test_when_endpoint_has_no_url(self):
            path = "http://foo.com"
            request = RequestNode({"path": path}, endpoint=EndpointNode({}))
            assert request.full_url_path == path

        def test_when_endpoint_has_url(self):
            endpoint_path = "http://foo.com/api"
            endpoint = EndpointNode({"path": endpoint_path})
            request = RequestNode({"path": "/foo"}, endpoint=endpoint)
            assert request.full_url_path == f"http://foo.com/api/foo"

        def test_with_trailing_slashes(self):
            endpoint = EndpointNode({"path": "http://foo.com/"})
            request = RequestNode({"path": "/foo/"}, endpoint=endpoint)
            assert request.full_url_path == "http://foo.com/foo/"

        def test_calls_evaluate(self, mocker, mock_evaluate):
            endpoint = EndpointNode({"path": "http://foo.com/"})
            request = RequestNode({"path": "/foo/"}, endpoint=endpoint)
            request.full_url_path
            calls = [mocker.call("http://foo.com/"), mocker.call("/foo/")]

            mock_evaluate.assert_has_calls(calls)

    class TestHeaders:
        def test_when_endpoint_has_no_headers(self):
            headers = {"abc": "def"}
            request = RequestNode({"headers": headers}, endpoint=EndpointNode({}))
            assert request.headers == headers

        def test_when_endpoint_has_headers(self):
            headers = {"abc": "def"}
            endpoint_headers = {"xxx": "www"}
            request = RequestNode(
                {"headers": headers},
                endpoint=EndpointNode({"headers": endpoint_headers}),
            )
            assert request.headers == {"abc": "def", "xxx": "www"}

        def test_with_repeated_keys(self):
            headers = {"abc": "def"}
            endpoint_headers = {"xxx": "www", "abc": "zxc"}
            request = RequestNode(
                {"headers": headers},
                endpoint=EndpointNode({"headers": endpoint_headers}),
            )
            assert request.headers == {"abc": "def", "xxx": "www"}

        def test_calls_evaluate(self, mocker, mock_evaluate):
            endpoint = EndpointNode({"headers": {"abc": "def"}})

            request = RequestNode({"headers": {"ghi": "jkl"}}, endpoint=endpoint)
            request.headers
            calls = [mocker.call({"abc": "def", "ghi": "jkl"})]

            mock_evaluate.assert_has_calls(calls)

    class TestParams:
        def test_when_endpoint_has_no_params(self):
            params = {"abc": "def"}
            request = RequestNode({"params": params}, endpoint=EndpointNode({}))
            assert request.params == params

        def test_when_endpoint_has_params(self):
            params = {"abc": "def"}
            endpoint_params = {"xxx": "www"}
            request = RequestNode(
                {"params": params}, endpoint=EndpointNode({"params": endpoint_params})
            )
            assert request.params == {"abc": "def", "xxx": "www"}

        def test_with_repeated_keys(self):
            params = {"abc": "def"}
            endpoint_params = {"xxx": "www", "abc": "zxc"}
            request = RequestNode(
                {"params": params}, endpoint=EndpointNode({"params": endpoint_params})
            )
            assert request.params == {"abc": "def", "xxx": "www"}

        def test_calls_evaluate(self, mocker, mock_evaluate):
            endpoint = EndpointNode({"params": {"abc": "def"}})

            request = RequestNode({"params": {"ghi": "jkl"}}, endpoint=endpoint)
            request.params
            calls = [mocker.call({"abc": "def", "ghi": "jkl"})]

            mock_evaluate.assert_has_calls(calls)

    class TestBody:
        def test_when_request_has_no_body(self):
            request = RequestNode({}, endpoint=EndpointNode({}))
            assert request.body == {}

        def test_when_request_has_no_body(self):
            request = RequestNode({"body": {"abc": "def"}}, endpoint=EndpointNode({}))
            assert request.body == {"abc": "def"}

        def test_calls_evaluate(self, mocker, mock_evaluate):
            request = RequestNode({"body": {"ghi": "jkl"}}, endpoint=EndpointNode({}))
            request.body
            calls = [mocker.call({"ghi": "jkl"})]

            mock_evaluate.assert_has_calls(calls)

    class TestRun:
        @pytest.fixture
        def mock_request(self, mocker):
            return mocker.patch("scanapi.tree.request_node.requests.request")

        def test_calls_request(self, mock_request):
            request = RequestNode({}, endpoint=EndpointNode({}))
            request.run()

            mock_request.assert_called_once_with(
                request.http_method,
                request.full_url_path,
                headers=request.headers,
                params=request.params,
                json=request.body,
                allow_redirects=False,
            )

    class TestValidate:
        @pytest.fixture()
        def mock_validate_keys(self, mocker):
            return mocker.patch("scanapi.tree.request_node.validate_keys")

        def test_should_call_validate_keys(self, mock_validate_keys):
            spec = {"headers": {"foo": "bar"}, "name": "foo", "path": "foo.bar"}
            node = RequestNode(spec, endpoint=EndpointNode({}))
            keys = spec.keys()
            node._validate()

            mock_validate_keys.assert_called_with(
                keys,
                ("body", "headers", "method", "name", "params", "path", "vars"),
                "request",
            )
            assert len(keys) == 3
            assert "headers" in keys
            assert "name" in keys
            assert "path" in keys



from scanapi.evaluators.code_evaluator import CodeEvaluator
from scanapi.evaluators.spec_evaluator import SpecEvaluator
from scanapi.evaluators.string_evaluator import StringEvaluator

__all__ = ["CodeEvaluator", "SpecEvaluator", "StringEvaluator"]



import os

from scanapi.config_loader import load_config_file

DEFAULT_CONFIG_PATH = ".scanapi.yaml"


class Settings(dict):
    def __init__(self):
        self["spec_path"] = "api.yaml"
        self["reporter"] = "markdown"
        self["output_path"] = None
        self["template"] = None

        super(dict, self).__init__()

    def save_config_file_preferences(self, config_path=None):
        if not config_path and not self.has_default_config_file:
            return

        if not config_path:
            user_config = load_config_file(DEFAULT_CONFIG_PATH)
            self["config_path"] = DEFAULT_CONFIG_PATH
            self.update(**user_config)
            return

        user_config = load_config_file(config_path)
        self.update(**user_config)

    def save_click_preferences(self, **preferences):
        cleaned_preferences = {k: v for k, v in preferences.items() if v is not None}
        self.update(**cleaned_preferences)

    def save_preferences(self, **click_preferences):
        config_path = (
            click_preferences["config_path"]
            if "config_path" in click_preferences
            else None
        )
        self.save_config_file_preferences(config_path)
        self.save_click_preferences(**click_preferences)

    @property
    def has_default_config_file(self):
        return os.path.isfile(DEFAULT_CONFIG_PATH)


settings = Settings()



import os
import pytest

from scanapi.errors import BadConfigurationError, InvalidPythonCodeError
from scanapi.evaluators.spec_evaluator import SpecEvaluator
from scanapi.tree import EndpointNode


class TestSpecEvaluator:
    @pytest.fixture
    def mock_string_evaluate(self, mocker):
        return mocker.patch(
            "scanapi.evaluators.spec_evaluator.StringEvaluator.evaluate"
        )

    @pytest.fixture
    def spec_evaluator(self):
        endpoint = EndpointNode({})
        return SpecEvaluator(endpoint)

    class TestEvaluateString:
        def test_should_call_evaluate_dict(self, spec_evaluator, mock_string_evaluate):
            string = "foo"
            spec_evaluator.evaluate(string)
            assert mock_string_evaluate.called_once_with(string)

    class TestEvaluateDict:
        class TestWhenDictIsEmpty:
            def test_return_empty_dict(self, spec_evaluator):
                evaluated_dict = spec_evaluator.evaluate({})
                assert len(evaluated_dict) == 0

        class TestWhenDictIsNotEmpty:
            def test_return_evaluated_dict(
                self, spec_evaluator, mocker, mock_string_evaluate
            ):
                mock_string_evaluate.side_effect = ["foo", "bar"]
                assert spec_evaluator.evaluate({"app_id": "foo", "token": "bar"}) == {
                    "app_id": "foo",
                    "token": "bar",
                }

                mock_string_evaluate.assert_has_calls(
                    [
                        mocker.call("foo", spec_evaluator),
                        mocker.call("bar", spec_evaluator),
                    ]
                )

    class TestEvaluateList:
        class TestWhenListIsEmpty:
            def test_return_empty_list(self, spec_evaluator):
                evaluated_list = spec_evaluator.evaluate([])
                assert len(evaluated_list) == 0

        class TestWhenListIsNotEmpty:
            def test_return_evaluated_list(
                self, spec_evaluator, mocker, mock_string_evaluate
            ):
                values = ["foo", "bar"]
                mock_string_evaluate.side_effect = values
                assert spec_evaluator.evaluate(values) == ["foo", "bar"]

                mock_string_evaluate.assert_has_calls(
                    [
                        mocker.call("foo", spec_evaluator),
                        mocker.call("bar", spec_evaluator),
                    ]
                )



from itertools import chain
import logging


from scanapi.evaluators import SpecEvaluator
from scanapi.tree.tree_keys import (
    ENDPOINTS_KEY,
    HEADERS_KEY,
    NAME_KEY,
    PARAMS,
    PATH_KEY,
    REQUESTS_KEY,
)
from scanapi.tree.request_node import RequestNode
from scanapi.utils import join_urls, validate_keys

logger = logging.getLogger(__name__)


class EndpointNode:
    SCOPE = "endpoint"
    ALLOWED_KEYS = (
        ENDPOINTS_KEY,
        HEADERS_KEY,
        NAME_KEY,
        PARAMS,
        PATH_KEY,
        REQUESTS_KEY,
    )

    def __init__(self, spec, parent=None):
        self.spec = spec
        self.parent = parent
        self.child_nodes = []
        self.__build()
        self.vars = SpecEvaluator(self, spec.get("vars", {}))

    def __build(self):
        self._validate()

        self.child_nodes = [
            EndpointNode(spec, parent=self) for spec in self.spec.get("endpoints", [])
        ]

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    @property
    def name(self):
        return self.spec.get("name", "root")

    @property
    def path(self):
        path = self.spec.get("path", "").strip()
        url = join_urls(self.parent.path, path) if self.parent else path

        return self.vars.evaluate(url)

    @property
    def headers(self):
        return self._get_specs("headers")

    @property
    def params(self):
        return self._get_specs("params")

    def run(self):
        for request in self._get_requests():
            try:
                yield request.run()
            except Exception as e:
                error_message = (
                    f"Error to make request `{request.full_url_path}`. {str(e)}"
                )
                logger.error(error_message)
                continue

    def _validate(self):
        validate_keys(self.spec.keys(), self.ALLOWED_KEYS, self.SCOPE)

    def _get_specs(self, field_name):
        values = self.spec.get(field_name, {})
        parent_values = getattr(self.parent, field_name, None)

        if parent_values:
            return {**parent_values, **values}

        return values

    def _get_requests(self):
        return chain(
            (RequestNode(spec, self) for spec in self.spec.get("requests", [])),
            *(child._get_requests() for child in self.child_nodes),
        )



import logging
import re

from scanapi.errors import InvalidPythonCodeError

logger = logging.getLogger(__name__)


class CodeEvaluator:
    python_code_pattern = re.compile(
        r"(?P<something_before>\w*)(?P<start>\${{)(?P<python_code>.*)(?P<end>}})(?P<something_after>\w*)"
    )  # ${{<python_code>}}

    @classmethod
    def evaluate(cls, sequence, vars):
        # To avoid circular imports
        from scanapi.evaluators.string_evaluator import StringEvaluator

        # Available imports to be used dinamically in the API spec
        import datetime
        import math
        import random
        import time
        import uuid

        match = cls.python_code_pattern.search(sequence)

        if not match:
            return sequence

        code = match.group("python_code")

        try:
            response = vars.get("response")
            python_code_value = str(eval(code))
            return StringEvaluator.replace_var_with_value(
                sequence, match.group(), python_code_value
            )
        except Exception as e:
            raise InvalidPythonCodeError(str(e))



import logging
import requests

from scanapi.errors import HTTPMethodNotAllowedError
from scanapi.evaluators import SpecEvaluator
from scanapi.tree.tree_keys import (
    BODY_KEY,
    HEADERS_KEY,
    METHOD_KEY,
    NAME_KEY,
    PARAMS,
    PATH_KEY,
    VARS_KEY,
)
from scanapi.utils import join_urls, hide_sensitive_info, validate_keys

logger = logging.getLogger(__name__)


class RequestNode:
    SCOPE = "request"
    ALLOWED_KEYS = (
        BODY_KEY,
        HEADERS_KEY,
        METHOD_KEY,
        NAME_KEY,
        PARAMS,
        PATH_KEY,
        VARS_KEY,
    )
    ALLOWED_HTTP_METHODS = ("GET", "POST", "PUT", "PATCH", "DELETE")

    def __init__(self, spec, endpoint):
        self.spec = spec
        self.endpoint = endpoint
        self._validate()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.full_url_path}>"

    def __getitem__(self, item):
        return self.spec[item]

    @property
    def http_method(self):
        method = self.spec.get("method", "get").upper()
        if method not in self.ALLOWED_HTTP_METHODS:
            raise HTTPMethodNotAllowedError(method, self.ALLOWED_HTTP_METHODS)

        return method

    @property
    def name(self):
        return self["name"]

    @property
    def full_url_path(self):
        base_path = self.endpoint.path
        path = self.spec.get("path", "")
        full_url = join_urls(base_path, path)

        return self.endpoint.vars.evaluate(full_url)

    @property
    def headers(self):
        endpoint_headers = self.endpoint.headers
        headers = self.spec.get("headers", {})

        return self.endpoint.vars.evaluate({**endpoint_headers, **headers})

    @property
    def params(self):
        endpoint_params = self.endpoint.params
        params = self.spec.get("params", {})

        return self.endpoint.vars.evaluate({**endpoint_params, **params})

    @property
    def body(self):
        body = self.spec.get("body", {})

        return self.endpoint.vars.evaluate(body)

    def run(self):
        method = self.http_method
        url = self.full_url_path
        logger.info("Making request %s %s", method, url)

        response = requests.request(
            method,
            url,
            headers=self.headers,
            params=self.params,
            json=self.body,
            allow_redirects=False,
        )

        self.endpoint.vars.update(
            self.spec.get("vars", {}), extras={"response": response}, preevaluate=True
        )

        hide_sensitive_info(response)
        return response

    def _validate(self):
        validate_keys(self.spec.keys(), self.ALLOWED_KEYS, self.SCOPE)



import os
import pytest

from scanapi.errors import BadConfigurationError
from scanapi.evaluators import StringEvaluator


class TestStringEvaluator:
    class TestEvaluate:
        @pytest.fixture
        def mock__evaluate_env_var(self, mocker):
            mock_func = mocker.patch(
                "scanapi.evaluators.string_evaluator.StringEvaluator._evaluate_env_var"
            )
            mock_func.return_value = ""
            return mock_func

        @pytest.fixture
        def mock_code_evaluate(self, mocker):
            mock_func = mocker.patch(
                "scanapi.evaluators.code_evaluator.CodeEvaluator.evaluate"
            )
            mock_func.return_value = ""
            return mock_func

        def test_calls_code_evaluate(self, mock_code_evaluate):
            StringEvaluator.evaluate("boo", {})
            mock_code_evaluate.assert_called_once_with("boo", {})

        def test_calls__evaluate_env_var(self, mock__evaluate_env_var):
            StringEvaluator.evaluate("boo", {})
            mock__evaluate_env_var.assert_called_once_with("boo")

    class TestEvaluateEnvVar:
        class TestWhenDoesNotMatchThePattern:
            test_data = ["no env var", "${var}", "${MyVar}", "${{var}}", "${{VAR}}"]

            @pytest.mark.parametrize("sequence", test_data)
            def test_should_return_sequence(self, sequence):
                assert StringEvaluator._evaluate_env_var(sequence) == sequence

        class TestWhenMatchesThePattern:
            class TestWhenEnvVarIsSetProperly:
                @pytest.fixture(autouse=True)
                def base_url_env(self):
                    os.environ["BASE_URL"] = "https://jsonplaceholder.typicode.com"
                    os.environ["POST_ID"] = "2"

                test_data = [
                    ("${BASE_URL}", "https://jsonplaceholder.typicode.com"),
                    ("${BASE_URL}/posts", "https://jsonplaceholder.typicode.com/posts"),
                    (
                        "https://jsonplaceholder.typicode.com/posts/${POST_ID}",
                        "https://jsonplaceholder.typicode.com/posts/2",
                    ),
                    (
                        "${BASE_URL}/posts/${POST_ID}",
                        "https://jsonplaceholder.typicode.com/posts/2",
                    ),
                ]

                @pytest.mark.parametrize("sequence, expected", test_data)
                def test_should_return_evaluated_var(self, sequence, expected):
                    assert StringEvaluator._evaluate_env_var(sequence) == expected

            class TestWhenThereIsNoCorrespondingEnvVar:
                @pytest.fixture(autouse=True)
                def remove_base_url_env(self):
                    if os.environ.get("BASE_URL"):
                        del os.environ["BASE_URL"]

                def test_should_raise_bad_configuration_error(self):
                    with pytest.raises(BadConfigurationError) as excinfo:
                        StringEvaluator._evaluate_env_var("${BASE_URL}")

                    assert (
                        str(excinfo.value)
                        == "'BASE_URL' environment variable not set or badly configured"
                    )

    class TestEvaluateCustomVar:
        class TestWhenDoesNotMatchThePattern:
            test_data = ["no var", "${ENV_VAR}", "${{code}}"]

            @pytest.mark.parametrize("sequence", test_data)
            def test_should_return_sequence(self, sequence):
                assert StringEvaluator._evaluate_custom_var(sequence, {}) == sequence

        class TestWhenMatchesThePattern:
            class TestWhenCodeDoesNotContainThePreSavedCustomVar:
                test_data = [
                    ("${user_id}"),
                    ("something before ${user_id} something after"),
                    ("something before ${user_id}"),
                    ("${user_id} something after"),
                ]

                @pytest.mark.parametrize("sequence", test_data)
                def test_should_return_sequence(self, sequence):
                    assert (
                        StringEvaluator._evaluate_custom_var(sequence, {}) == sequence
                    )

            class TestWhenCodeContainsThePreSavedCustomVar:
                test_data = [
                    ("${user_id}", "10"),
                    ("${apiKey}", "abc123"),
                    ("something before ${user_id}", "something before 10"),
                    (
                        "something before ${user_id} something after",
                        "something before 10 something after",
                    ),
                    ("${user_id} something after", "10 something after"),
                ]

                @pytest.mark.parametrize("sequence, expected", test_data)
                def test_should_return_sequence(self, sequence, expected):
                    vars = {"user_id": "10", "apiKey": "abc123"}
                    assert (
                        StringEvaluator._evaluate_custom_var(sequence, vars) == expected
                    )

    class TestReplaceVarWithValue:
        test_data = [
            ("${age}", "${age}", "45", "45"),
            ("I am ${age} years old", "${age}", "60", "I am 60 years old"),
            (
                "url/${{some['python_code']!}}",
                "${{some['python_code']!}}",
                "300",
                "url/300",
            ),
            (
                "${BASE_URL}/posts",
                "${BASE_URL}",
                "https://jsonplaceholder.typicode.com",
                "https://jsonplaceholder.typicode.com/posts",
            ),
        ]

        @pytest.mark.parametrize(
            "sequence, variable, variable_value, expected", test_data
        )
        def test_should_replace(self, sequence, variable, variable_value, expected):
            assert (
                StringEvaluator.replace_var_with_value(
                    sequence, variable, variable_value
                )
                == expected
            )



import setuptools

requirements = [
    "click >= 7.0",
    "curlify >= 2.2.1",
    "jinja2 >= 2.10.3",
    "pyyaml >= 5.1.1",
    "requests >= 2.22.0",
]
dev_requirements = ["black >= 19.10b0", "ipdb >= 0.13.2", "pre-commit >= 1.20.0"]
test_requirements = [
    "codecov >= 2.0.15",
    "pytest >= 5.2.4",
    "pytest-cov >= 2.8.1",
    "pytest-mock >= 1.11.2",
    "requests-mock >= 1.7.0",
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scanapi",
    version="0.0.19",
    author="Camila Maia",
    author_email="cmaiacd@gmail.com",
    description="Automated Testing and Documentation for your REST API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/scanapi/scanapi",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    extras_require={"dev": dev_requirements, "test": test_requirements},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points="""
        [console_scripts]
        scanapi = scanapi:scan
    """,
    zip_safe=False,
    include_package_data=True,
    package_data={"scanapi": ["scanapi/templates/*"]},
)



"""
Code based on solution https://gist.github.com/joshbode/569627ced3076931b02f
"""

from typing import Any, IO
import json
import logging
import os
import yaml

from scanapi.errors import EmptyConfigFileError, FileFormatNotSupportedError

logger = logging.getLogger(__name__)


class Loader(yaml.SafeLoader):
    """YAML/JSON Loader with `!include` constructor."""

    def __init__(self, stream: IO) -> None:
        """Initialise Loader."""

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        super().__init__(stream)


def construct_include(loader: Loader, node: yaml.Node) -> Any:
    """Include file referenced at node."""

    relative_path = os.path.join(loader._root, loader.construct_scalar(node))
    full_path = os.path.abspath(relative_path)
    extension = os.path.splitext(full_path)[1].lstrip(".")

    with open(full_path, "r") as f:
        if extension in ("yaml", "yml"):
            return yaml.load(f, Loader)
        elif extension in ("json",):
            return json.load(f)
        else:
            raise FileFormatNotSupportedError(f".{extension}", relative_path)


def load_config_file(file_path):
    extension = os.path.splitext(file_path)[-1]

    if extension not in (".yaml", ".yml", ".json"):
        raise FileFormatNotSupportedError(extension, file_path)

    with open(file_path, "r") as stream:
        logger.info(f"Loading file {file_path}")
        data = yaml.load(stream, Loader)

        if not data:
            raise EmptyConfigFileError(file_path)

        return data


yaml.add_constructor("!include", construct_include, Loader)



import pytest

from scanapi.tree import EndpointNode


class TestEndpointNode:
    @pytest.fixture
    def empty_node(self):
        return EndpointNode({})

    class TestInit:
        def test_should_create_children(self):
            endpoints = [{}, {}]
            node = EndpointNode({"endpoints": endpoints})
            assert len(node.child_nodes) == len(endpoints)

    class TestName:
        def test_when_node_is_empty(self, empty_node):
            assert empty_node.name == "root"

    class TestPath:
        @pytest.fixture
        def mock_evaluate(self, mocker):
            mock_func = mocker.patch(
                "scanapi.tree.endpoint_node.SpecEvaluator.evaluate"
            )
            mock_func.return_value = ""

            return mock_func

        def test_when_node_is_empty(self, empty_node):
            assert empty_node.path == ""

        def test_when_parent_has_no_url(self):
            base_path = "http://foo.com"
            node = EndpointNode({"path": base_path}, parent=EndpointNode({}))
            assert node.path == base_path

        def test_when_parent_has_url(self):
            base_path = "http://foo.com/api"
            parent = EndpointNode({"path": base_path})
            node = EndpointNode({"path": "/foo"}, parent=parent)
            assert node.path == f"http://foo.com/api/foo"

        def test_with_trailing_slashes(self):
            parent = EndpointNode({"path": "http://foo.com/"})
            node = EndpointNode({"path": "/foo/"}, parent=parent)
            assert node.path == "http://foo.com/foo/"

        def test_calls_evaluate(self, mocker, mock_evaluate):
            parent = EndpointNode({"path": "http://foo.com/"})
            node = EndpointNode({"path": "/foo/"}, parent=parent)
            node.path
            calls = [mocker.call("http://foo.com/"), mocker.call("/foo/")]

            mock_evaluate.assert_has_calls(calls)

    class TestHeaders:
        def test_when_node_is_empty(self, empty_node):
            assert empty_node.headers == {}

        def test_when_parent_has_no_headers(self):
            headers = {"abc": "def"}
            node = EndpointNode({"headers": headers}, parent=EndpointNode({}))
            assert node.headers == headers

        def test_when_parent_has_headers(self):
            headers = {"abc": "def"}
            parent_headers = {"xxx": "www"}
            node = EndpointNode(
                {"headers": headers}, parent=EndpointNode({"headers": parent_headers})
            )
            assert node.headers == {"abc": "def", "xxx": "www"}

        def test_when_parent_has_repeated_keys(self):
            headers = {"abc": "def"}
            parent_headers = {"xxx": "www", "abc": "zxc"}
            node = EndpointNode(
                {"headers": headers}, parent=EndpointNode({"headers": parent_headers})
            )
            assert node.headers == {"abc": "def", "xxx": "www"}

    class TestParams:
        def test_when_node_is_empty(self, empty_node):
            assert empty_node.params == {}

        def test_when_parent_has_no_params(self):
            params = {"abc": "def"}
            node = EndpointNode({"params": params}, parent=EndpointNode({}))
            assert node.params == params

        def test_when_parent_has_params(self):
            params = {"abc": "def"}
            parent_params = {"xxx": "www"}
            node = EndpointNode(
                {"params": params}, parent=EndpointNode({"params": parent_params})
            )
            assert node.params == {"abc": "def", "xxx": "www"}

        def test_when_parent_has_repeated_keys(self):
            params = {"abc": "def"}
            parent_params = {"xxx": "www", "abc": "zxc"}
            node = EndpointNode(
                {"params": params}, parent=EndpointNode({"params": parent_params})
            )
            assert node.params == {"abc": "def", "xxx": "www"}

    class TestRun:
        pass  # TODO

    class TestValidate:
        @pytest.fixture()
        def mock_validate_keys(self, mocker):
            return mocker.patch("scanapi.tree.endpoint_node.validate_keys")

        def test_should_call_validate_keys(self, mock_validate_keys):
            spec = {"headers": {"foo": "bar"}, "name": "foo", "path": "foo.bar"}
            node = EndpointNode(spec)
            keys = spec.keys()
            node._validate()

            mock_validate_keys.assert_called_with(
                keys,
                ("endpoints", "headers", "name", "params", "path", "requests"),
                "endpoint",
            )
            assert len(keys) == 3
            assert "headers" in keys
            assert "name" in keys
            assert "path" in keys

    class TestGetSpecs:
        pass  # TODO

    class TestGetRequests:
        def test_when_node_is_empty(self, empty_node):
            requests = list(empty_node._get_requests())
            assert len(requests) == 0

        def test_when_node_has_children(self):
            node = EndpointNode(
                {
                    "endpoints": [
                        {
                            "name": "foo",
                            "requests": [{"name": "First"}, {"name": "Second"}],
                        }
                    ]
                }
            )
            requests = list(node._get_requests())
            assert len(requests) == 2



from scanapi.errors import InvalidKeyError
from scanapi.settings import settings

ALLOWED_ATTRS_TO_HIDE = ("headers body").split()


def join_urls(first_url, second_url):
    if not first_url:
        return second_url

    if not second_url:
        return first_url

    first_url = first_url.strip("/")
    second_url = second_url.lstrip("/")

    return "/".join([first_url, second_url])


def validate_keys(keys, available_keys, scope):
    for key in keys:
        if not key in available_keys:
            raise InvalidKeyError(key, scope, available_keys)


def hide_sensitive_info(response):
    report_settings = settings.get("report", {})
    request = response.request
    request_settings = report_settings.get("hide-request", {})
    response_settings = report_settings.get("hide-response", {})

    _hide(request, request_settings)
    _hide(response, response_settings)


def _hide(http_msg, hide_settings):
    for http_attr in hide_settings:
        secret_fields = hide_settings[http_attr]
        for field in secret_fields:
            _override_info(http_msg, http_attr, field)


def _override_info(http_msg, http_attr, secret_field):
    if (
        secret_field in getattr(http_msg, http_attr)
        and http_attr in ALLOWED_ATTRS_TO_HIDE
    ):
        getattr(http_msg, http_attr)[secret_field] = "<sensitive_information>"
